import matplotlib.pyplot as plt
import random
import time
import the_env


from stable_baselines3 import PPO, TD3
#model = PPO.load("ppo_best_model_same_weights")
#model = PPO.load("ppo_best_model_diffr_weights")
#model = PPO.load("ppo_best_model_diffr_weights_2nd")
#model = TD3.load("epidemic_td3")

# initial values ---------------------------------------------
epidemic_data = [14, 0.1] #incubation coef and fatality coef |
demographic_data = [100000, 3, 1000] #N, I, density           |
# ------------------------------------------------------------
custom_data = [epidemic_data, demographic_data]
env = the_env.MorEpiEnv(custom_data)

env.a_weights[0] = 1 #weight on A1
env.a_weights[1] = 1 #weight on A2
env.a_weights[2] = 1 #weight on A3
env.a_weights[3] = 1 #weight on A4
env.a_weights[4] = 1 #weight on A5
env.a_weights[5] = 1 #weight on A6
env.a_weights[6] = 1 #weight on A7

# to enable vaccination usage by the agent
# env.set_vaccination(True):

num_ep = 1

reward_avrg_list = []

action_list = []
for ep in range(num_ep):
    obs = env.reset()
    step_reward_list = []
    done = False
    while not done:
        #action, _ = model.predict(obs) #, deterministic=False)
        # or
        action = env.action_space.sample()

        obs, reward, done, info = env.step(action)

        env.render('IRD') # scores or graph or IRD
        
        #step_reward_list.append(reward)

    #print(len(step_reward_list))
    #print("episode score = ", sum(step_reward_list)/50)

##        if done:
##            print("episode: ", ep,  "finished at : ", step+1)
##            reward_avrg = sum(step_reward_list) / len(step_reward_list)
##            print("with average reward= ", reward_avrg)
##            
##    reward_avrg_list.append(reward_avrg)

env.close()

##import matplotlib.pyplot as plt
##fig2, ax2 = plt.subplots(nrows=1, ncols=1)
##plt.plot(s2s)
##fig2.show()






# with scores plot
import networkx as nx 
from matplotlib.animation import FuncAnimation
import matplotlib.pyplot as plt
import numpy as np
import random

global ax, fig, G, l1, l2, l3, l4, l5, l6, l7, l8, edgelist, pos, e_pos
fig, ax = plt.subplots(figsize=(10,7))
G = nx.DiGraph()
l1, l2, l3, l4, l5, l6, l7, l8, l9 ='Cur_I', 'N', 'Frst_I', 'Re_I', 'Nxt_I', 'U', 'K', 'R', 'D'
edgelist= [(l1,l2),(l2,l3),(l2,l4),(l3,l5),(l5, l6),(l5, l7),(l7, l8),(l7, l9), (l6, l1), (l8,l2)]

pos = {l1: (10, 80), l2: (10, 50), l3:(60, 50), l4:(10, 20),  #I, N, FI, RI
       l5:(100, 50), l6:(150, 80), l7:(150, 20), l8:(100,20), l9:(200, 20) #NI, U, K, R, D
       }

e_pos = {l1: (10, 80), l2: (10, 50), l3:(60, 50), l4:(10, 20),  #I, N, FI, RI
       l5:(100, 50), l6:(150, 80), l7:(150, 20), l8:(100,20), l9:(200, 20) #NI, U, K, R, D
       }
def showNet(ep, step, states, quantities, h_score, e_score, reward, R0, show, save):
    global ax, fig, G, l1, l2, l3, l4, l5, l6, l7, l8, I9, edgelist, pos, e_pos

    #plt.style.use('fivethirtyeight')
    #plt.style.use('Solarize_Light2')
    
    fig.clf()            
        
    s1 = float("{:.3f}".format(states[0]))
    s2 = float("{:.3f}".format(states[1]))
    s3 = float("{:.3f}".format(states[2]))
    s4 = float("{:.3f}".format(states[3]))
    
    e_labels = [s1, s4, float("{:.3f}".format(1-s2)), s2, float("{:.3f}".format(1-s3)), s3]
    
    edge_labels = {(l2,l3):e_labels[0],(l2,l4):e_labels[1], (l5,l6):e_labels[2],
                   (l5, l7):e_labels[3],(l7,l8):e_labels[4], (l7,l9):e_labels[5]}

    nx.draw_networkx_nodes(G, pos=pos, nodelist=list(pos.keys()), node_size=2000, alpha=1)
                                                                        #put here, connectionstyle="arc3,rad=0.1" for curved edges
    nx.draw_networkx_edges(G, pos=e_pos, edgelist= edgelist , edge_color="gray",
                           arrows=True, arrowsize=20, node_size=2000, alpha=1) #arrowstyle='-|>' or 'wedge'

    nx.draw_networkx_edge_labels(G,pos,edge_labels=edge_labels, font_size=10) 
    nx.draw_networkx_labels(G, pos=pos, labels=dict(zip(pos.keys(),pos.keys())),  font_color="black") 
    
    xx = list(pos.values())
    # Cur_I
    x,y= xx[0]
    popul_Cur_I = int(quantities[0])
    plt.text(x,y-4.6,s= popul_Cur_I, fontsize=10, bbox=dict(facecolor='yellow', alpha=0.6),horizontalalignment='center')

    # N
    x,y= xx[1]
    popul_N = int(quantities[1])
    plt.text(x,y-4.6,s= popul_N, fontsize=10, bbox=dict(facecolor='yellow', alpha=0.6),horizontalalignment='center')

    # F_I
    x,y= xx[2]
    popul_FI = int(quantities[2])
    plt.text(x,y-4.6,s= popul_FI, fontsize=10, bbox=dict(facecolor='yellow', alpha=0.6),horizontalalignment='center')

    # Re_I
    x,y= xx[3]
    popul_ReI = int(quantities[3])
    plt.text(x,y-4.6,s= popul_ReI, fontsize=10, bbox=dict(facecolor='yellow', alpha=0.6),horizontalalignment='center')

    # Nxt_I
    x,y= xx[4]
    popul_NI = int(quantities[4])
    plt.text(x,y-4.6,s= popul_NI, fontsize=10, bbox=dict(facecolor='yellow', alpha=0.6),horizontalalignment='center')
    
    # U
    x,y= xx[5]
    popul_U = int(quantities[5])
    plt.text(x,y-4.6,s= popul_U, fontsize=10, bbox=dict(facecolor='yellow', alpha=0.6),horizontalalignment='center')

    # K
    x,y= xx[6]
    popul_K = int(quantities[6])
    plt.text(x,y-4.6,s= popul_K, fontsize=10, bbox=dict(facecolor='yellow', alpha=0.6),horizontalalignment='center')

    # R
    x,y= xx[7]
    popul_R = int(quantities[7])
    plt.text(x,y-4.6,s= popul_R, fontsize=10, bbox=dict(facecolor='yellow', alpha=0.6),horizontalalignment='center')

    # D
    x,y= xx[8]
    popul_D = int(quantities[8])
    plt.text(x,y-4.6,s= popul_D, fontsize=10, bbox=dict(facecolor='yellow', alpha=0.6),horizontalalignment='center')

    R0 = float("{:.3f}".format(R0))
    h_score = "{:.3f}".format(h_score)
    e_score = "{:.3f}".format(e_score)
    
    plt.title('episode= '+str(ep)+ ' | step= '+str(step) + ' | health score= '+str(h_score) +
              ' | economy score= '+str(e_score) + ' | R0= '+str(R0) + ' | reward= '+str(reward), 
              ha="center", va="bottom", size=10, color="orange") 

    # Hide grid lines
    plt.grid(False)

    if save==True:
        return fig

    if show==True:
        plt.ion()
        fig.show() 
        plt.pause(0.0000001)

    else:
        pass

# Test:
##gg = showNet(10, 44, [11, 22, 33], [1, 2, 3, 4, 5, 6, 7],
##                    0.44, 0.55, show=False, save=True)


        





import matplotlib.pyplot as plt
import numpy as np
import random
import gym
from gym import spaces

from . graph import showNet
from . scores import plot_scores
from . IRD import plot_ird

class MorEpiEnv(gym.Env): 
    def __init__(self, custom_data):
        self.observation_space = spaces.Box(
            low= 0,
            high= 1,
            shape=(4,),
            dtype=np.float32
        )
        self.action_space = spaces.Box(
            low= -1,
            high= 1,
            shape=(7,),
            dtype=np.float32
        )

        print('Initializing the env ...')
        # exactly the reset function
        self.epidemic_data = custom_data[0]
        self.demographic_data = custom_data[1]
        
        I, N = self.demographic_data[1], self.demographic_data[0]
        FI, ReI, Nxt_I, U, K, R, D = 1, 0, 1, 1, 1, 1, 0
        self.quantities = [I, N, FI, ReI, Nxt_I, U, K, R, D]

        self.R0 = 2
        #self.R0 = np.interp(self.R0, [1.4, 2.4], [0.01, 0.8])    # R0 forced between 0.1 and 0.8

        self.TI_TD = [25, 2]
        self.Te = 0
        self.a_weights = [1, 1, 1, 1, 1, 1, 1]  
        self.vaccination = False
        self.St = 0
        
        self.reward = 0
        self.h_score = 0
        self.e_score = 1
        self.h_w_sum = 0
        self.e_w_sum = 0

        self.ep = 0
        self.steps = 0
        
        a = np.random.rand(7)
        Density = self.demographic_data[2]
        
        s1 = (1 - (a[0]+ a[1]+ a[2]+ a[3] - self.St)/4) * Density * (self.R0 + (self.quantities[3]/(self.quantities[1]+0.00001)))
        s1 = s1 - (s1*a[6]*0.92)
        s2 = self.epidemic_data[0] * a[4]
        s3 = self.epidemic_data[1] * a[5]
        s4 = (1 - a[6]) * 0.46
        
        self.states = np.array([s1, s2, s3, s4], dtype=np.float32)
        self.actions = a
        self.done = False

        self.actions_list = []
        
        print('env successfully initialized ======_======= \n')
        
    def set_TI_TD(self, TI_TD):
        self.TI_TD = TI_TD

    def set_Te(self, Te):
        self.Te = Te
        
    def set_a_weights(self, wts):
        self.a_weights[0] = wts[0] 
        self.a_weights[1] = wts[1] 
        self.a_weights[2] = wts[2]
        self.a_weights[3] = wts[3]
        self.a_weights[4] = wts[4]
        self.a_weights[5] = wts[5]
        self.a_weights[6] = wts[6]


    def set_vaccination(self, vaccine):
        self.vaccination = vaccine

    def set_St(self, st):
        self.st = st

    def set_R0(self, R0):
        #self.R0 = np.interp(R0, [1.4, 2.4], [0.01, 0.8])    # This is what's declared in the paper
        self.R0 = R0  # This is what's used in the validation and in PYPI

    def reset(self):
        print('Reseting the env ...')
        
        I, N = self.demographic_data[1], self.demographic_data[0]
        FI, ReI, Nxt_I, U, K, R, D = 1, 0, 1, 1, 1, 1, 0
        self.quantities = [I, N, FI, ReI, Nxt_I, U, K, R, D]

        self.R0 = 2
        #self.R0 = np.interp(self.R0, [1.4, 2.4], [0.01, 0.8])    # R0 forced between 0.1 and 0.8
        

        self.TI_TD = [25, 2]
        self.Te = 0
        self.a_weights = [1, 1, 1, 1, 1, 1, 1]  
        self.vaccination = False
        
        self.reward = 0
        self.h_score = 0
        self.e_score = 1
        self.h_w_sum = 0
        self.e_w_sum = 0

        self.ep = 0
        self.steps = 0
        
        a = np.random.rand(7)
        Density = self.demographic_data[2]
        
        s1 = (1 - (a[0]+ a[1]+ a[2]+ a[3]- self.St)/4) * Density * (self.R0 + (self.quantities[3]/(self.quantities[1]+0.00001)))
        s1 = s1 - (s1*a[6]*0.92)
        s2 = self.epidemic_data[0] * a[4]
        s3 = self.epidemic_data[1] * a[5]
        s4 = (1 - a[6]) * 0.46
        
        self.states = np.array([s1, s2, s3, s4], dtype=np.float32)
        self.actions = a
        self.done = False

        self.actions_list = []
        print('Env reset')
              
        return self.states

    def step(self, action):
        # actions of the model are between -1 and 1
        a = action
        # remap actions to be between 0 and 1,
        # add small random value to prevent 0 and for little stochasticity
        a = np.interp(a, [-1, 1], [0, 1]) 

        self.actions_list.append(a) 
                                             
        Density = self.demographic_data[2]
        Density = np.interp(Density, [3, 1000000], [0.02, 0.8])  # density forced between 0.1 and 0.8
        IN = self.quantities[0]/(self.quantities[1]+0.000001)  # IN is certainly between 0 and 1

        incubation = self.epidemic_data[0]
        incubation = np.interp(incubation, [5.6, 12.5], [0.1, 0.9]) #incubation forced between 0.1 and 0.9

        fatality = self.epidemic_data[1] + random.uniform(0, 0.08)

        if self.vaccination == True:
            pass
        else:
            a[6] = 0 #no vaccination

        nxt_s1 = (1 - ((a[0]+ a[1]+ a[2]+ a[3]- self.St)/4)) * Density * (self.R0+IN)
        nxt_s1 = nxt_s1 - (nxt_s1*a[6]*0.92)
        nxt_s2 = (1 - incubation) * a[4]
        nxt_s3 = (1 - a[5]) * fatality
        nxt_s4 = (1 - a[6]) * 0.08

        self.states = np.array([nxt_s1, nxt_s2, nxt_s3, nxt_s4], dtype=np.float32)
        
        #below is : I = I + U          
        I = self.quantities[0] + self.quantities[5]
        #below is : N= N - Nxt_I + R
        N = self.quantities[1] - self.quantities[4] + self.quantities[7]
        FI = int(N* nxt_s1)
        ReI = int(nxt_s4*self.quantities[7])
        Nxt_I = FI + ReI
        U = int(Nxt_I*(1-nxt_s2))
        K = int(Nxt_I*nxt_s2)
        R = int(K*(1-nxt_s3))
        D = int(K*nxt_s3)
        self.quantities = [I, N, FI, ReI, Nxt_I, U, K, R, D]

        ww = self.a_weights
        if self.vaccination == False:
            ww[6] = 0
        self.e_score = 1 - ((ww[0]*a[0]+ ww[1]*a[1]+ ww[2]*a[2]+ ww[3]*a[3]+ ww[4]*a[4]+ ww[5]*a[5] + ww[6]*a[6])/sum(ww))

        if Nxt_I < self.TI_TD[0] and D < self.TI_TD[1]:
            self.h_score = (self.TI_TD[0] - Nxt_I) / self.TI_TD[0]
        else:
            self.h_score = 0
        
        if self.h_score > 0 and self.e_score > self.Te: 
            self.reward = self.e_score + self.h_score
        else:
            self.reward = 0.00000


        self.reward = float("{:.4f}".format(self.reward))

        self.steps = self.steps + 1

        info = {}
        if self.steps == 360:  #360days #50 for tests
            self.done = True
            self.ep = self.ep + 1
            print('h_score, e_score : ', self.h_score, self.e_score)
            print('s1, s2, s3, s4 : ', nxt_s1, nxt_s2, nxt_s3, nxt_s4)
            print('last reward : ', self.reward)
            print('TI, TD, Te, St : ', self.TI_TD[0], self.TI_TD[1], self.Te, self.St)
            print('a weights: ', self.a_weights)
            print('vaccination: ', self.vaccination)
            print("\n -------- episode: ", self.ep,  " is finished ----------- \n")
            return self.states, self.reward, self.done, info
        else:
            #print('R0 : ', self.R0)
            #print('N : ', self.quantities[1])
            return self.states, self.reward, self.done, info


    def render(self, mode=None, save=False):
        if mode=='scores' or mode== None:
            if save==True:
                plt.close(1)
                plt.close(3)
                figgg = plot_scores(self.h_score, self.e_score, self.ep,
                            self.steps, self.reward, self.R0, self.Te, show=not save, save=save)
                return figgg
            if save==False:
                plt.close(1)
                plt.close(3)
                plot_scores(self.h_score, self.e_score, self.ep,
                            self.steps, self.reward,  self.R0, self.Te, show=not save, save=save)

        if mode=='IRD':
            if save==True:
                plt.close(1)
                plt.close(2)
                figgg = plot_ird(self.quantities[4], self.quantities[7],
                            self.quantities[8], self.ep, self.steps, self.reward,
                                 self.R0, show=not save, save=save)
                return figgg
            if save==False:
                plt.close(1)
                plt.close(2)
                plot_ird(self.quantities[4], self.quantities[7],
                            self.quantities[8], self.ep, self.steps, self.reward,
                                 self.R0, show=not save, save=save)


        if mode=='graph':
            if save==True:
                plt.close(2)
                plt.close(3)
                figgg = showNet(self.ep, self.steps, self.states, self.quantities,
                        self.h_score, self.e_score, self.reward,  self.R0, show=not save, save=save)
                return figgg
            
            if save==False:
                plt.close(2)
                plt.close(3)
                showNet(self.ep, self.steps, self.states, self.quantities,
                        self.h_score, self.e_score, self.reward,  self.R0, show=not save, save=save)
        else:
            pass
    
    def close(self):
        plt.close('all')

    def interpret_policy(self, show, k=20): #k = w: window size
        w = k
        aaa = self.actions_list
        a1 = []
        a2 = []
        a3 = []
        a4 = []
        a5 = []
        a6 = []
        a7 = []
        for x in aaa:
            a1.append(x[0])
            a2.append(x[1])
            a3.append(x[2])
            a4.append(x[3])
            a5.append(x[4])
            a6.append(x[5])
            a7.append(x[6])
            
        interpreted1 = []
        interpreted2 = []
        interpreted3 = []
        interpreted4 = []
        interpreted5 = []
        interpreted6 = []
        interpreted7 = []
        for i in range(int(len(a1)/w)):
            x1 = a1[i*w: (i+1)*w]
            interpreted1.append(np.mean(x1))  
            x2 = a2[i*w: (i+1)*w]
            interpreted2.append(np.mean(x2))
            x3 = a3[i*w: (i+1)*w]
            interpreted3.append(np.mean(x3))
            x4 = a4[i*w: (i+1)*w]
            interpreted4.append(np.mean(x4))
            x5 = a5[i*w: (i+1)*w]
            interpreted5.append(np.mean(x5))
            x6 = a6[i*w: (i+1)*w]
            interpreted6.append(np.mean(x6))
            x7 = a7[i*w: (i+1)*w]
            interpreted7.append(np.mean(x7))

        interpreted = [interpreted1, interpreted2, interpreted3,
                       interpreted4, interpreted5, interpreted6,
                       interpreted7]

        if show==True:
            # get the last sequence of actions with k avrerage
            j = 1
            for i in interpreted:
                plt.subplot(7,1,j)
                plt.plot(i)
                j = j+1
            plt.show()
        else:
            return interpreted
            
        



        






import matplotlib.pyplot as plt
import numpy as np
import random


global fig2, ax1, ax2, h_scores, e_scores
fig2 = plt.figure(figsize=(12,6))
ax1 = fig2.add_subplot(211)
ax2 = fig2.add_subplot(212)

h_scores = []
e_scores = []

def plot_scores(h_score, e_score, ep, step, reward, R0, Te, show, save):
    global fig2, ax1, ax2, h_scores, e_scores


    plt.style.use('fivethirtyeight')
    #plt.style.use('Solarize_Light2')

    if  h_score > -1: 
        h_scores.append(h_score)
    else:
        h_scores.append(-0.1)
    ax1.cla()
    #ax1.axhline(y=0.5, color='y', linestyle='--')
    ax1.axvline(x=(ep)*50, color='black', linestyle='--')
    ax1.axvline(x=(ep+1)*50, color='black', linestyle='--')
    ax1.plot (h_scores, 'r.-', linewidth=0.5)

    R0 = float("{:.3f}".format(R0))
    h_score = float("{:.3f}".format(h_score))
    ax1.set_title('episode= '+str(ep)+ ' | step= '+str(step) +
                  ' | health score = '+str(h_score)+ ' | R0= '+str(R0)
                  + ' | reward= '+str(reward),
              ha="center", va="bottom", size=10, color="orange") 

    e_scores.append(e_score) 
    ax2.cla()
    ax2.axhline(y=Te, color='y', linestyle='--')
    ax2.axvline(x=(ep)*50, color='black', linestyle='--')
    ax2.axvline(x=(ep+1)*50, color='black', linestyle='--')
    ax2.plot (e_scores, 'b.-', linewidth=0.5)
    e_score = float("{:.3f}".format(e_score))
    ax2.set_title('episode= '+str(ep)+ ' | step= '+str(step) +
                  ' | economy score= '+str(e_score) + ' | R0= '+str(R0)
                  + ' | reward= '+str(reward),
              ha="center", va="bottom", size=10, color="orange") 

    if show==True:
        plt.ion()
        fig2.show() 
        plt.pause(0.0000001)
    if save==True:
        return fig2
    else:
        pass







import matplotlib.pyplot as plt
import numpy as np
import random


global fig3, ax1, Is, Rs, Ds
fig3 = plt.figure(figsize=(9,6))
ax1 = fig3.add_subplot(111)

Is = []
Rs = []
Ds = []
def plot_ird(I, R, D, ep, step, reward, R0, show, save):
    global fig2, ax1, Is, Rs, Ds

    plt.style.use('fivethirtyeight')

    Is.append(I)
    Rs.append(R)
    Ds.append(D)
    
    ax1.cla()

    #uncoment thus if u want to update the viewed frame interval
    #plt.xlim(ep*50, (ep+2)*50)

    ax1.axvline(x=(ep)*50, color='black', linestyle='--', linewidth=0.8)
    ax1.axvline(x=(ep+1)*50, color='black', linestyle='--', linewidth=0.8)

    ax1.plot (Is, 'r.-', linewidth=0.5, label="Infected")
    ax1.plot (Rs, 'g.-', linewidth=0.5, label="Recovered")
    ax1.plot (Ds, 'b.-', linewidth=0.5, label="Deaths")

    plt.legend(loc='upper left')

    R0 = float("{:.3f}".format(R0))
    ax1.set_title('episode= '+str(ep)+ ' | step= '+str(step) +' | R0= '+str(R0)
                  + ' | reward= '+str(reward),
              ha="center", va="bottom", size=10, color="orange") 

    if show==True:
        plt.ion()
        fig3.show() 
        plt.pause(0.0000001)
    if save==True:
        return fig3
    else:
        pass






from setuptools import setup, find_packages
 
classifiers = [
  'Development Status :: 5 - Production/Stable',
  'Intended Audience :: Education',
  'Operating System :: Microsoft :: Windows :: Windows 10',
  'License :: OSI Approved :: MIT License',
  'Programming Language :: Python :: 3'
]

setup(
  name='MorEpiSim',
  version='0.0.2',
  description='a Reinforcement Learning based Epidemic control simulation environment',
  long_description=open('README.rst').read(),
  url='',  
  author='M.A.CHADI',
  author_email='m.aminechadi@gmail.com',
  license='MIT', 
  classifiers=classifiers,
  keywords='RL_based_Epidemic_control', 
  packages=["MorEpiSim"],
  include_package_data=True,
  #packages= find_packages(),
  install_requires=['numpy', 'matplotlib', 'networkx', 'gym']
                      # and openCV, imageIO : for play_and_save_video()
)
